import sys
import argparse

from docstring_parser import DocstringStyle

from sophia_doc import ModuleNode
from sophia_doc.utils import import_module
from sophia_doc.builders.markdown import MarkdownBuilder

if sys.version_info >= (3, 9):
    from argparse import BooleanOptionalAction
else:
    from argparse import Action

    class BooleanOptionalAction(Action):
        def __init__(
            self,
            option_strings,
            dest,
            default=None,
            type=None,
            choices=None,
            required=False,
            help=None,
            metavar=None,
        ):

            _option_strings = []
            for option_string in option_strings:
                _option_strings.append(option_string)

                if option_string.startswith("--"):
                    option_string = "--no-" + option_string[2:]
                    _option_strings.append(option_string)

            if help is not None and default is not None:
                help += " (default: %(default)s)"

            super().__init__(
                option_strings=_option_strings,
                dest=dest,
                nargs=0,
                default=default,
                type=type,
                choices=choices,
                required=required,
                help=help,
                metavar=metavar,
            )

        def __call__(self, parser, namespace, values, option_string=None):
            if option_string in self.option_strings:
                setattr(namespace, self.dest, not option_string.startswith("--no-"))

        def format_usage(self):
            return " | ".join(self.option_strings)


parser = argparse.ArgumentParser(
    description="Sophia_doc is a python package to automatically "
                "generate API documents for Python modules",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
)
parser.add_argument("module", type=str, help="Python module names.")
parser.add_argument(
    "-o",
    "--output-dir",
    type=str,
    default="doc",
    help="The directory to write document.",
)
parser.add_argument(
    "--docstring-style",
    type=str,
    default="auto",
    help="Docstring style the python module used.",
)
parser.add_argument(
    "--ignore-data",
    type=bool,
    action=BooleanOptionalAction,
    default=False,
    help="Ignore data in Markdown text.",
)
parser.add_argument(
    "--anchor-extend",
    type=bool,
    action=BooleanOptionalAction,
    default=False,
    help="Add anchor to markdown title.",
)
parser.add_argument(
    "--overwrite",
    type=bool,
    action=BooleanOptionalAction,
    default=False,
    help="Overwrite any file in output directory.",
)
parser.add_argument(
    "--exclude-module-name",
    type=bool,
    action=BooleanOptionalAction,
    default=False,
    help="Write file to path which exclude module name.",
)
parser.add_argument(
    "--init-file-name",
    type=str,
    default="index.md",
    help="The name of Markdown file from __init__.py, index.md by default.",
)


def cli():
    args = parser.parse_args()
    MarkdownBuilder(
        ModuleNode(import_module(args.module)),
        docstring_style=getattr(DocstringStyle, args.docstring_style.upper()),
        anchor_extend=args.anchor_extend,
        ignore_data=args.ignore_data,
    ).write(
        args.output_dir,
        overwrite=args.overwrite,
        exclude_module_name=args.exclude_module_name,
        init_file_name=args.init_file_name,
    )


if __name__ == "__main__":
    cli()



"""Builder is class to build ModuleNode to target formats."""
from pathlib import Path
from typing import TYPE_CHECKING
from abc import ABC, abstractmethod

from docstring_parser import Docstring, DocstringStyle, parse

if TYPE_CHECKING:
    from sophia_doc import DocNode, ModuleNode


class Builder(ABC):
    """Base class of Builder.

    Builds target formats from ModuleNode.

    Attributes:
        module: A ModuleNode object.
        docstring_style: The docstring style the module used, auto check by default.
    """

    module: "ModuleNode"
    docstring_style: DocstringStyle

    def __init__(
        self,
        module: "ModuleNode",
        *,
        docstring_style: DocstringStyle = DocstringStyle.AUTO,
    ):
        self.module = module
        self.docstring_style = docstring_style

    def _new_builder(self, module: "ModuleNode") -> "Builder":
        """Get a new instance of Builder class, is used in write method."""
        return self.__class__(module, docstring_style=self.docstring_style)

    def get_docstring(self, obj: "DocNode") -> "Docstring":
        """Get the Docstring object of a DocNode object.

        Args:
            obj: A DocNode object.

        Returns:
            A Docstring object.
        """
        return parse(obj.docstring, style=self.docstring_style)

    @abstractmethod
    def text(self) -> str:
        """Get string of current module documentation."""
        raise NotImplementedError

    @abstractmethod
    def get_path(self, **kwargs) -> Path:
        """Get the path to write file."""
        raise NotImplementedError

    def write(self, output_dir: str, *, overwrite: bool = False, **kwargs) -> None:
        """Write file to output dir.

        Args:
            output_dir: The output directory.
            overwrite: If true will overwrite any file in output directory,
                otherwise raise an Exception when file already exists.
        """
        filepath = Path(output_dir).resolve() / self.get_path(**kwargs)
        filepath.parent.mkdir(parents=True, exist_ok=True)
        if not self.module.is_namespace:
            filepath.touch(exist_ok=overwrite)
            with open(filepath, "w", encoding="utf-8") as f:
                f.write(self.text())
        for submodule in self.module.submodules:
            self._new_builder(submodule).write(
                output_dir, overwrite=overwrite, **kwargs
            )



"""Sophia_doc is a tool to automatically generate API documents for Python modules.

Typical usage example:

```
import sophia_doc
import sophia_doc.builders.markdown

module = sophia_doc.ModuleNode(sophia_doc)
builder = sophia_doc.builders.markdown.MarkdownBuilder(module)
builder.write('doc_dir')
```
"""
import sys
import inspect
import pkgutil
import warnings
import traceback
from enum import Enum
from functools import cached_property
from types import MethodType, ModuleType, FunctionType, MethodDescriptorType
from typing import Any, Dict, List, Tuple, Union, Generic, TypeVar, Optional, NamedTuple

from sophia_doc.utils import import_module

_T = TypeVar("_T")


def _find_class(func: Any) -> Optional[type]:
    # cut from pydoc
    cls = sys.modules.get(func.__module__)
    if cls is None:
        return None
    for name in func.__qualname__.split(".")[:-1]:
        cls = getattr(cls, name)
    if not inspect.isclass(cls):
        return None
    return cls


def _find_doc(obj: Any) -> Optional[str]:
    # cut from pydoc
    if inspect.ismethod(obj):
        name = obj.__func__.__name__
        self = obj.__self__
        if (
            inspect.isclass(self)
            and getattr(getattr(self, name, None), "__func__") is obj.__func__
        ):
            # class method
            cls = self
        else:
            cls = self.__class__
    elif inspect.isfunction(obj):
        name = obj.__name__
        cls = _find_class(obj)
        if cls is None or getattr(cls, name) is not obj:
            return None
    elif inspect.isbuiltin(obj):
        name = obj.__name__
        self = obj.__self__
        if inspect.isclass(self) and self.__qualname__ + "." + name == obj.__qualname__:
            # class method
            cls = self
        else:
            cls = self.__class__
    # Should be tested before isdatadescriptor().
    elif isinstance(obj, property):
        func = obj.fget
        name = func.__name__
        cls = _find_class(func)
        if cls is None or getattr(cls, name) is not obj:
            return None
    elif inspect.ismethoddescriptor(obj) or inspect.isdatadescriptor(obj):
        name = obj.__name__
        cls = obj.__objclass__
        if getattr(cls, name) is not obj:
            return None
        if inspect.ismemberdescriptor(obj):
            slots = getattr(cls, "__slots__", None)
            if isinstance(slots, dict) and name in slots:
                return slots[name]
    else:
        return None
    for base in cls.__mro__:
        try:
            doc = _get_own_doc(getattr(base, name))
        except AttributeError:
            continue
        if doc is not None:
            return doc
    return None


def _get_own_doc(obj: Any) -> Optional[str]:
    """Get the documentation string for an object
    if it is not inherited from its class.
    """
    # cut from pydoc
    try:
        doc = object.__getattribute__(obj, "__doc__")
        if doc is None:
            return None
        if obj is not type:
            typedoc = type(obj).__doc__
            if isinstance(typedoc, str) and typedoc == doc:
                return None
        return doc
    except AttributeError:
        return None


def _getdoc(obj: Any):
    """Get the documentation string for an object.

    All tabs are expanded to space. To clean up docstrings that are
    indented to line up with blocks of code, any whitespace than can be
    uniformly removed from the second line onwards is removed.
    """
    # cut from pydoc
    doc = _get_own_doc(obj)
    if doc is None:
        try:
            doc = _find_doc(obj)
        except (AttributeError, TypeError):
            return None
    if not isinstance(doc, str):
        return None
    return inspect.cleandoc(doc)


def getdoc(obj: Any) -> str:
    """Get the docstring string or comments for an object."""
    # cut from pydoc
    result = _getdoc(obj) or inspect.getcomments(obj)
    return result or ""


def isdata(obj: object) -> bool:
    """Check if an object is of a type that probably means it's data."""
    # cut from pydoc
    return not (
        inspect.ismodule(obj)
        or inspect.isclass(obj)
        or inspect.isroutine(obj)
        or inspect.isframe(obj)
        or inspect.istraceback(obj)
        or inspect.iscode(obj)
    )


def is_visible_name(name: str, _all: Optional[list] = None) -> bool:
    """Decide whether to show documentation on a variable.

    Args:
        name: The variable name.
        _all: __all__ list of a module.

    Returns:
        Show documentation on a variable or not.
    """
    if name == "__init__":
        return True
    # only document that which the programmer exported in __all__
    if _all is not None:
        return name in _all
    else:
        return not name.startswith("_")


def get_annotations(obj: Any) -> Dict[str, Any]:
    """Get the annotations dict for an object."""
    # refs: https://docs.python.org/3/howto/annotations.html
    if sys.version_info >= (3, 10):
        return inspect.get_annotations(obj)
    else:
        if isinstance(obj, type):
            return obj.__dict__.get("__annotations__", {})
        else:
            return getattr(obj, "__annotations__", {})


class DocNode(Generic[_T]):
    """The base class of document node.

    Attributes:
        obj: An object.
        name: The name of this object.
        module: The module of this object.
    """

    __slots__ = ("obj", "name", "module", "_qualname")
    obj: _T
    name: str
    module: "ModuleNode"
    _qualname: str

    def __init__(self, obj: _T, name: str, qualname: str, module: "ModuleNode"):
        self.obj = obj
        self.name = name
        self._qualname = qualname
        self.module = module

    @cached_property
    def qualname(self) -> str:
        if hasattr(self.obj, "__qualname__"):
            return self.obj.__qualname__
        return self._qualname

    @cached_property
    def realname(self) -> str:
        """Real name of this object."""
        return getattr(self.obj, "__name__", self.name)

    @cached_property
    def annotations(self) -> Dict[str, Any]:
        """Annotations of this object."""
        return get_annotations(self.obj)

    @cached_property
    def docstring(self) -> str:
        """Docstring of this object."""
        return getdoc(self.obj)

    @staticmethod
    def from_obj(obj: Any, name: str, qualname: str, module: "ModuleNode") -> "DocNode":
        """Returns an object of DocNode's subclass.

        Args:
            obj: An object.
            name: The name of the object.
            qualname: The qualname of the object.
            module: The module of the object.

        Returns:
            A DocNode object.
        """
        try:
            if inspect.ismodule(obj):
                return ModuleNode(obj)
            if inspect.isclass(obj):
                return ClassNode(obj, name, qualname, module)
            if inspect.isroutine(obj):
                return FunctionNode(obj, name, qualname, module)
        except AttributeError:
            pass
        if isdata(obj):
            return DataNode(obj, name, qualname, module)
        return OtherNode(obj, name, qualname, module)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}:{self.name} "
            f"obj:{self.obj} docstring:{self.docstring}"
        )


class ModuleNode(DocNode[ModuleType]):
    """The class of module node."""

    def __init__(self, obj: ModuleType):
        super().__init__(obj, obj.__name__, "", self)

    @cached_property
    def attributes(self) -> List[DocNode]:
        """A list of attributes of this module."""
        _all = getattr(self.obj, "__all__", None)
        attributes = []
        for key, value in list(getattr(self.obj, "__dict__", {}).items()):
            if _all is not None or (inspect.getmodule(value) or self.obj) is self.obj:
                if is_visible_name(key, _all):
                    attributes.append(self.from_obj(value, key, key, self))
        return attributes

    @cached_property
    def submodules(self) -> List["ModuleNode"]:
        """A list of submodules of this module."""
        submodules = []
        submodule_names = set()
        if self.is_package:
            for importer, modname, ispkg in pkgutil.iter_modules(self.obj.__path__):
                if not is_visible_name(modname):
                    continue
                try:
                    submodule_names.add(modname)
                    module = import_module(self.name + "." + modname)
                    submodules.append(ModuleNode(module))
                except ImportError:
                    warnings.warn(
                        f"Can not import {modname}:\n{traceback.format_exc()}"
                    )

        for key, value in inspect.getmembers(self.obj, inspect.ismodule):
            if (
                value.__name__.startswith(self.name + ".")
                and key not in submodule_names
            ):
                submodules.append(ModuleNode(value))
        return submodules

    @cached_property
    def file(self) -> Optional[str]:
        """The source file name of this module."""
        try:
            return inspect.getabsfile(self.obj)
        except TypeError:
            return None

    @cached_property
    def source(self) -> Optional[str]:
        """The source of this module."""
        loader = getattr(self.obj, "__loader__", None)
        if loader and getattr(loader, "get_source", None):
            try:
                source = loader.get_source(self.name)
                if source:
                    return source
            except ImportError:
                pass
        try:
            return inspect.getsource(self.obj)
        except OSError:
            return None

    @cached_property
    def is_package(self) -> bool:
        """Returns True if this module is a package."""
        return hasattr(self.obj, "__path__")

    @cached_property
    def is_namespace(self) -> bool:
        """Returns True if this module is a namespace package."""
        return hasattr(self.obj, "__path__") and not hasattr(self.obj, "__file__")

    @cached_property
    def classes(self) -> List["ClassNode"]:
        """A list of class objects in this module's attributes."""
        return [i for i in self.attributes if isinstance(i, ClassNode)]

    @cached_property
    def functions(self) -> List["FunctionNode"]:
        """A list of function objects in this module's attributes."""
        return [i for i in self.attributes if isinstance(i, FunctionNode)]

    @cached_property
    def data(self) -> List["DataNode"]:
        """A list of data objects in module's this attributes."""
        return [i for i in self.attributes if isinstance(i, DataNode)]


class ClassNode(DocNode[type]):
    """The class of class node."""

    class Attribute(NamedTuple):
        name: str
        kind: str
        node: DocNode

    @cached_property
    def attributes(self) -> List["ClassNode.Attribute"]:
        """A list of attributes of this class."""
        attributes = []
        for name, kind, cls, value in inspect.classify_class_attrs(self.obj):
            if (is_visible_name(name) and cls is self.obj) or (
                name == "__init__" and not isinstance(self.obj, Enum)
            ):
                if inspect.isdatadescriptor(value):
                    # ignore data descriptor create by __slots__
                    if name in getattr(self.obj, "__slots__", []):
                        continue
                    kind = "data descriptor"
                    if isinstance(value, property) and value.fset is None:
                        kind = "readonly property"
                # get original function from class method or static method
                if kind == "class method" or kind == "static method":
                    value = value.__func__
                # functools.cached_property needs special handling
                if isinstance(value, cached_property):
                    kind = "cached property"
                    node = DataNode(
                        value, name, self.qualname + "." + name, self.module
                    )

                else:
                    node = DocNode.from_obj(
                        value, name, self.qualname + "." + name, self.module
                    )
                attributes.append(self.Attribute(name, kind, node))
        return attributes

    @cached_property
    def subclasses(self) -> List["ClassNode"]:
        """A list of subclasses of this class."""
        return [
            ClassNode(cls, cls.__name__, cls.__qualname__, self.module)
            for cls in type.__subclasses__(self.obj)
            if not (cls.__name__.startswith("_") and cls.__module__ == "builtins")
        ]

    @cached_property
    def bases(self) -> Tuple[str, ...]:
        """Base class names of this class."""
        return tuple(
            map(
                lambda x: (x.__module__ + "." if x.__module__ != "builtins" else "")
                + x.__qualname__,
                self.obj.__bases__,
            )
        )

    @cached_property
    def mro(self) -> Tuple[type, ...]:
        """The mro of this class."""
        return inspect.getmro(self.obj)

    @cached_property
    def is_abstract(self) -> bool:
        """Returns True if this class is an abstract class."""
        return inspect.isabstract(self.obj)


class FunctionNode(DocNode[Union[FunctionType, MethodType, MethodDescriptorType]]):
    """The class of function node."""

    @cached_property
    def signature(self) -> Optional[inspect.Signature]:
        """Signature of this function."""
        try:
            return inspect.signature(self.obj)
        except (ValueError, TypeError):
            return None

    @cached_property
    def is_async(self) -> bool:
        """Returns True if this function is an async function."""
        return inspect.iscoroutinefunction(self.obj) or inspect.isasyncgenfunction(
            self.obj
        )

    @cached_property
    def is_bound_method(self) -> bool:
        """Returns True if this function is a bound method."""
        if inspect.ismethod(self.obj):
            return True
        if inspect.isbuiltin(self.obj):
            _self = getattr(self.obj, "__self__", None)
            return not (inspect.ismodule(_self) or (_self is None))
        return False

    @cached_property
    def is_lambda_func(self):
        """Returns True if this function is a lambda function."""
        return self.realname == "<lambda>"


class DataNode(DocNode[_T]):
    """The class of data node."""

    @cached_property
    def annotations(self) -> Dict[str, Any]:
        """Annotations of this object."""
        if isinstance(self.obj, property):
            return get_annotations(self.obj.fget)
        if isinstance(self.obj, cached_property):
            return get_annotations(self.obj.func)
        return get_annotations(self.obj)


class OtherNode(DocNode[_T]):
    """The class of other node."""

    pass



"""Utils of sophia_doc."""
import re
import inspect
import warnings
import importlib
import traceback
from types import ModuleType
from typing import Any, Optional


def import_module(modname: str) -> ModuleType:
    """A wrapper of importlib.import_module, convert exceptions to ImportError.

    Args:
        modname: The name of the module to be imported.

    Returns:
        The imported module.
    """
    try:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=ImportWarning)
            return importlib.import_module(modname)
    except BaseException as exc:
        if isinstance(exc, KeyboardInterrupt):
            # Dot not catch KeyboardInterrupt.
            # Catch KeyboardInterrupt may prevent user kill python
            # when the module took a too long time to import.
            raise exc
        raise ImportError(exc, traceback.format_exc()) from exc


def format_annotation(annotation: Any, base_module: Optional[ModuleType] = None) -> str:
    """Format the annotation.

    Args:
        annotation: The annotation object to be formatted.
        base_module: A module which the object from.

    Returns:
        The formatted annotation.
    """
    if annotation is inspect.Signature.empty:
        return ""
    if isinstance(annotation, str):
        return annotation
    # use regex delete 'ForwardRef' from annotation
    return re.sub(
        r"\bForwardRef\((?P<quot>[\'\"])(?P<string>.*?)(?P=quot)\)",
        r"\g<string>",
        inspect.formatannotation(annotation, base_module),
    )


def format_parameter(parameter: inspect.Parameter, type_comments: bool = False) -> str:
    """Format inspect.Parameter object, type comments is optional.

    Cut from inspect.Parameter.__str__().

    Args:
        parameter: The inspect.Parameter object to be formatted.
        type_comments: If True will include type comments.

    Returns:
        The formatted string.
    """
    kind = parameter.kind
    formatted = parameter.name

    # Add annotation and default value
    if parameter.annotation is not inspect.Parameter.empty and type_comments:
        formatted = "{}: {}".format(formatted, format_annotation(parameter.annotation))

    if parameter.default is not inspect.Parameter.empty:
        if parameter.annotation is not inspect.Parameter.empty:
            formatted = "{} = {}".format(formatted, repr(parameter.default))
        else:
            formatted = "{}={}".format(formatted, repr(parameter.default))

    if kind == inspect.Parameter.VAR_POSITIONAL:
        formatted = "*" + formatted
    elif kind == inspect.Parameter.VAR_KEYWORD:
        formatted = "**" + formatted

    return formatted


def format_signature(signature: inspect.Signature, type_comments: bool = False) -> str:
    """
    Format inspect.Signature object, type comments is optional.

    Cut from inspect.Signature.__str__().

    Args:
        signature: The inspect.Signature object to be formatted.
        type_comments: If True will include type comments.

    Returns:
        The formatted string.
    """
    if type_comments:
        return str(signature)

    result = []
    render_pos_only_separator = False
    render_kw_only_separator = True
    for param in signature.parameters.values():
        formatted = format_parameter(param, type_comments=type_comments)

        kind = param.kind

        if kind == inspect.Parameter.POSITIONAL_ONLY:
            render_pos_only_separator = True
        elif render_pos_only_separator:
            # It's not a positional-only parameter, and the flag
            # is set to 'True' (there were pos-only params before.)
            result.append("/")
            render_pos_only_separator = False

        if kind == inspect.Parameter.VAR_POSITIONAL:
            # OK, we have an '*args'-like parameter, so we won't need
            # a '*' to separate keyword-only arguments
            render_kw_only_separator = False
        elif kind == inspect.Parameter.KEYWORD_ONLY and render_kw_only_separator:
            # We have a keyword-only parameter to render, and we haven't
            # rendered an '*args'-like parameter before, so add a '*'
            # separator to the parameters list ("foo(arg1, *, arg2)" case)
            result.append("*")
            # This condition should be only triggered once, so
            # reset the flag
            render_kw_only_separator = False

        result.append(formatted)

    if render_pos_only_separator:
        # There were only positional-only parameters, hence the
        # flag was not reset to 'False'
        result.append("/")

    rendered = "({})".format(", ".join(result))

    if signature.return_annotation is not inspect.Parameter.empty and type_comments:
        anno = format_annotation(signature.return_annotation)
        rendered += " -> {}".format(anno)

    return rendered



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sophia_doc', 'sophia_doc.builders']

package_data = \
{'': ['*']}

install_requires = \
['docstring-parser>=0.15,<0.16']

entry_points = \
{'console_scripts': ['sophia-doc = sophia_doc.__main__:cli']}

setup_kwargs = {
    'name': 'sophia-doc',
    'version': '0.1.1',
    'description': 'A python package to automatically generate API documents for Python modules.',
    'long_description': '# Sophia-doc\n\n**A python package to automatically generate API documents for Python modules.**\n\n## Introduction\n\nSophia is a python package to automatically generate API documents for Python modules.\n\nIt\'s a lot like sphinx, but it only focuses on generating markdown documentation.\n\nIt does not support PEP 224 attribute docstrings, because the PEP was rejected, and have to use ast module to support\nit, which brings additional complexity to this project.\n\n# Install\n\n```shell\npip install sophia-doc\n```\n\n# Quickstart\n\n```shell\nsophia_doc "sophia_doc" -o ./doc\n```\n\n## Usage\n\nCommand line:\n\n```shell\nusage: sophia_doc [-h] [-o OUTPUT_DIR] [--docstring-style DOCSTRING_STYLE] [--anchor-extend | --no-anchor-extend] [--overwrite | --no-overwrite]\n                   [--exclude-module-name | --no-exclude-module-name]\n                   module\n\nSophia_doc is a python package to automatically generate API documents for Python modules\n\npositional arguments:\n  module                Python module names.\n\noptions:\n  -h, --help            show this help message and exit\n  -o OUTPUT_DIR, --output-dir OUTPUT_DIR\n                        The directory to write document. (default: doc)\n  --docstring-style DOCSTRING_STYLE\n                        Docstring style the python module used. (default: auto)\n  --anchor-extend, --no-anchor-extend\n                        Add anchor to markdown title. (default: False)\n  --overwrite, --no-overwrite\n                        Overwrite any file in output directory. (default: False)\n  --exclude-module-name, --no-exclude-module-name\n                        Write file to path which exclude module name. (default: False)\n```\n\n## License\n\nMIT © st1020',
    'author': 'st1020',
    'author_email': 'stone_1020@qq.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import inspect
import warnings
from pathlib import Path
from textwrap import indent
from typing import Dict, List, Tuple, Optional

from docstring_parser import Docstring, DocstringParam, DocstringStyle

from sophia_doc.builders import Builder
from sophia_doc.utils import format_signature, format_annotation
from sophia_doc import DocNode, DataNode, ClassNode, ModuleNode, FunctionNode


def get_description(docstring: "Docstring") -> List[str]:
    """Get description form a Docstring object.

    Args:
        docstring: A Docstring object.

    Returns:
        A list of description string.
    """
    result: List[str] = list()
    if docstring.short_description:
        result.append(docstring.short_description)
    if docstring.long_description:
        result.append(docstring.long_description)
    return result


def parser_param(
    name: str, annotation: Optional[str], description: Optional[str]
) -> str:
    """Get formatted string of parameter.

    Args:
        name: Parameter name.
        annotation: Parameter annotation.
        description: Parameter description.

    Returns:
        Formatted string of parameter.
    """
    result = f"- {Markdown.bold(name)}"
    if annotation:
        result += f" ({Markdown.italic(annotation)})"
    if description:
        result += f" - {description}"
    return result


def parser_docstring_param(param: "DocstringParam") -> str:
    """Get formatted string of parameter from a DocstringParam object.

    Args:
        param: A DocstringParam object.

    Returns:
        Formatted string of parameter.
    """
    return parser_param(param.arg_name, param.type_name, param.description)


class Markdown:
    """Markdown format."""

    @staticmethod
    def indent(text: str, level: int = 1) -> str:
        return indent(text, prefix=" " * (level * 2))

    @staticmethod
    def italic(text: str) -> str:
        return f"*{text}*"

    @staticmethod
    def bold(text: str) -> str:
        return f"**{text}**"

    @staticmethod
    def title(text: str, level: int = 1) -> str:
        return "#" * level + " " + text

    @staticmethod
    def inline_code(text: str) -> str:
        return f"`{text}`"


class MarkdownBuilder(Builder):
    """
    Markdown Builder.

    Attributes:
        anchor_extend: If true will add anchor extend to title,
            like # Your title {#your-custom-id}
        ignore_data: If true will ignore data in Markdown text.
    """

    anchor_extend: bool
    ignore_data: bool

    def __init__(
        self,
        module: "ModuleNode",
        *,
        docstring_style: DocstringStyle = DocstringStyle.AUTO,
        anchor_extend: bool = False,
        ignore_data: bool = False,
    ):
        super().__init__(module, docstring_style=docstring_style)
        self.anchor_extend = anchor_extend
        self.ignore_data = ignore_data

    def _new_builder(self, module: "ModuleNode") -> "Builder":
        return self.__class__(
            module,
            docstring_style=self.docstring_style,
            anchor_extend=self.anchor_extend,
            ignore_data=self.ignore_data,
        )

    def get_path(
        self,
        exclude_module_name: bool = False,
        init_file_name: str = "index.md",
        **kwargs,
    ) -> Path:
        """Get the path to write file.

        Args:
            exclude_module_name: If true will write file to path
                which exclude module name, like change output path
                form `./doc/sophia_doc/index.md` to `./doc/index.md`.
            init_file_name: The name of Markdown file
                from __init__.py, `index.md` by default.
        """
        if exclude_module_name:
            path = Path(*self.module.name.split(".")[1:])
        else:
            path = Path(*self.module.name.split(".")[0:])
        return (
            path / init_file_name if self.module.is_package else path.with_suffix(".md")
        )

    def text(self) -> str:
        result: List[str] = list()
        result.append(Markdown.title(self.module.name, 1))

        docstring = self.get_docstring(self.module)
        result.extend(get_description(docstring))

        for node in self.module.attributes:
            result.append(self.build_doc(node))

        return self._build_str(result).replace("<", r"\<").replace(">", r"\>")

    @staticmethod
    def _build_str(str_list: List[str]) -> str:
        return "\n\n".join(filter(lambda x: x, str_list))

    def build_doc(self, node: "DocNode", *, level: int = 1, **kwargs) -> str:
        """Build markdown string from a DocNode.

        Args:
            node: A DocNode.
            level: The title level.

        Returns:
            A markdown string.
        """
        if isinstance(node, ClassNode):
            return self.build_class(node, level=level)
        elif isinstance(node, FunctionNode):
            return self.build_function(node, level=level, **kwargs)
        elif isinstance(node, DataNode):
            return self.build_data(node, level=level, **kwargs)

    def _extend_title(self, title: str, node: "DocNode") -> str:
        return title + " {#" + node.qualname + "}" if self.anchor_extend else title

    def build_class(self, node: ClassNode, *, level: int = 1) -> str:
        """Build markdown string from a ClassNode.

        Args:
            node: A ClassNode.
            level: The title level.

        Returns:
            A markdown string.
        """
        _kind = list()
        if node.is_abstract:
            _kind.append("abstract")
        if isinstance(node.obj, Exception):
            _kind.append("exception")
        else:
            _kind.append("class")
        _kind = " ".join(_kind)

        init = None
        for attr in node.attributes:
            if attr.name == "__init__":
                assert isinstance(attr.node, FunctionNode)
                init = attr.node

        result: List[str] = list()
        title = Markdown.title(
            Markdown.italic(_kind) + " " + Markdown.inline_code(node.name), level + 1
        )
        if init and init.signature:
            title += format_signature(init.signature)
        else:
            warnings.warn(f"Can not get __init__ method signature of class {node.name}")
        result.append(self._extend_title(title, node))

        result.append("Bases: " + ", ".join(map(Markdown.inline_code, node.bases)))

        docstring = self.get_docstring(node)
        result.extend(get_description(docstring))

        if init and init.signature:
            result.extend(
                self._build_argument(
                    init, self.get_docstring(init), ignore_first_arg=True
                )
            )

        if docstring.params or node.annotations:
            result.append("- **Attributes**")

            parma_dict: Dict[
                str, Tuple["inspect.Parameter", Optional["DocstringParam"]]
            ] = {}
            if node.annotations:
                parma_dict = {
                    key: (annotation, None)
                    for key, annotation in node.annotations.items()
                    if not key.startswith("_")
                }

            if docstring.params:
                for param_doc in docstring.params:
                    annotation, _ = parma_dict.get(param_doc.arg_name, (None, None))
                    if param_doc.type_name is None and annotation:
                        param_doc.type_name = format_annotation(
                            annotation, base_module=node.module.obj
                        )
                    parma_dict[param_doc.arg_name] = (annotation, param_doc)

            for name, (annotation, param_doc) in parma_dict.items():
                if param_doc is None:
                    result.append(
                        Markdown.indent(
                            parser_param(
                                name,
                                format_annotation(
                                    annotation, base_module=node.module.obj
                                ),
                                None,
                            )
                        )
                    )
                else:
                    result.append(Markdown.indent(parser_docstring_param(param_doc)))

        if docstring.examples:
            result.append("- **Examples**")
            result.append(Markdown.indent(docstring.examples[0].description))

        for name, kind, node in node.attributes:
            if name == "__init__":
                continue
            result.append(
                self.build_doc(
                    node,
                    level=level + 1,
                    kind=kind,
                    ignore_first_arg=kind == "method" or kind == "class method",
                )
            )

        return self._build_str(result)

    @staticmethod
    def _build_argument(
        node: FunctionNode, docstring: "Docstring", *, ignore_first_arg: bool = False
    ) -> List[str]:
        result = []
        if docstring.params or (node.signature and node.signature.parameters):
            parma_dict: Dict[
                str, Tuple["inspect.Parameter", Optional["DocstringParam"]]
            ] = {}
            if node.signature and node.signature.parameters:
                parma_dict = {
                    key: (param, None)
                    for key, param in node.signature.parameters.items()
                    if not (
                        param.kind == param.VAR_POSITIONAL
                        or param.kind == param.VAR_KEYWORD
                    )
                }

            if docstring.params:
                for param_doc in docstring.params:
                    if param_doc.arg_name not in parma_dict and (
                        node.signature and node.signature.parameters
                    ):
                        if not param_doc.arg_name.startswith("*"):
                            warnings.warn(
                                f'The argument "{param_doc.arg_name}" '
                                f"can not find in function signature."
                            )
                    param, _ = parma_dict.get(param_doc.arg_name, (None, None))
                    if param_doc.type_name is None and param:
                        param_doc.type_name = format_annotation(
                            param.annotation, base_module=node.module.obj
                        )
                    parma_dict[param_doc.arg_name] = (param, param_doc)

            if ignore_first_arg and parma_dict:
                parma_dict.pop(list(parma_dict.keys())[0])

            if parma_dict:
                result.append("- **Arguments**")

            for param, param_doc in parma_dict.values():
                if param_doc is None:
                    result.append(
                        Markdown.indent(
                            parser_param(
                                param.name,
                                format_annotation(
                                    param.annotation, base_module=node.module.obj
                                ),
                                None,
                            )
                        )
                    )
                else:
                    result.append(Markdown.indent(parser_docstring_param(param_doc)))

        return result

    def build_function(
        self,
        node: FunctionNode,
        *,
        level: int = 1,
        kind: str = "function",
        ignore_first_arg: bool = False,
        **kwargs,  # noqa
    ) -> str:
        """Build markdown string from a FunctionNode.

        Args:
            node: A FunctionNode.
            level: The title level.
            kind: The function kind, like 'function', 'method', 'class method'.
            ignore_first_arg: If True the first argument of the function
                will be ignored.

        Returns:
            A markdown string.
        """
        if not node.signature:
            warnings.warn(
                f"The {node.qualname} ({node.obj}) not have signature, ignored."
            )
            return ""

        _kind = list()
        if node.is_async:
            _kind.append("async")
        if node.is_lambda_func:
            _kind.append("lambda")
        _kind.append(kind)
        _kind = " ".join(_kind)

        result: List[str] = list()
        result.append(
            self._extend_title(
                Markdown.title(
                    Markdown.italic(_kind)
                    + " "
                    + Markdown.inline_code(
                        "{name}{signature}".format(
                            name=node.name, signature=format_signature(node.signature)
                        )
                    ),
                    level + 1,
                ),
                node,
            )
        )

        docstring = self.get_docstring(node)
        result.extend(get_description(docstring))

        result.extend(
            self._build_argument(node, docstring, ignore_first_arg=ignore_first_arg)
        )

        if docstring.returns or (
            node.signature
            and node.signature.return_annotation is not inspect.Signature.empty
        ):
            result.append("- **Returns**")

            type_name = ""
            if docstring.returns is not None and docstring.returns.type_name:
                type_name = docstring.returns.type_name
            elif (
                node.signature is not None
                and node.signature.return_annotation is not inspect.Signature.empty
            ):
                type_name = format_annotation(
                    node.signature.return_annotation, base_module=node.module.obj
                )

            if type_name:
                result.append(Markdown.indent(f"Type: {Markdown.italic(type_name)}"))

            if docstring.returns and docstring.returns.description:
                result.append(Markdown.indent(docstring.returns.description))

        if docstring.raises:
            result.append("- **Raises**")
            for raise_doc in docstring.raises:
                result.append(
                    Markdown.indent(
                        "- {type_name} - {description}".format(
                            type_name=Markdown.bold(raise_doc.type_name),
                            description=raise_doc.description,
                        )
                    )
                )

        if docstring.examples:
            result.append("- **Examples**")
            result.append(Markdown.indent(docstring.examples[0].description))

        return self._build_str(result)

    def build_data(
        self, node: DataNode, *, level: int = 1, kind: str = "data", **kwargs  # noqa
    ) -> str:
        """Build markdown string from a DataNode.

        Args:
            node: A DataNode.
            level: The title level.
            kind: The function kind, like 'data', 'property'.

        Returns:
            A markdown string.
        """
        if self.ignore_data and kind == "data":
            return ""

        result: List[str] = list()
        result.append(
            self._extend_title(
                Markdown.title(
                    Markdown.italic(kind) + " " + Markdown.inline_code(node.name),
                    level + 1,
                ),
                node,
            )
        )

        if "property" in kind and node.annotations.get("return", None):
            result.append(
                "Type: {type_name}".format(
                    type_name=Markdown.italic(
                        format_annotation(
                            node.annotations["return"], base_module=node.module.obj
                        )
                    )
                )
            )

        docstring = self.get_docstring(node)
        result.extend(get_description(docstring))

        return self._build_str(result)
